unit fpcompatibility;

{---  $define Linux  for Kylix Compiler}
{---  $define fpc    for Free Pascal - Compiler}
{---  $define clx    for Delphi 6 CLX - Projekte}

{$ifdef Linux} {$define CLX} {$endif}

{$define CLX}

interface

{********** Interface uses **********}


                         {********** all compilers **********}

{$ifndef fpc}            {********** Compiler Delphi oder Kylix **********}
{$ifndef Linux}          {********** Compiler Delphi ********* }
{$ifdef clx}             {********** CLX ********* }
uses SysUtils, classes, qextctrls, windows, SyncObjs;
{$else} {VCL}
uses SysUtils, classes, extctrls, windows, SyncObjs;
{$endif} {CLX}
{$else} {Linux}          {********** Compiler Kylix ********** }
uses SysUtils, classes, qextctrls, SyncObjs;
{$endif} {Linux}
{$else} {fpc}            {********** Compiler Free Pascal ********** }
uses classes, ipc;
{$endif} {fpc}

{********** end Interface uses **********}
                                               



{********** Interface definitions **********}     


                         {********** all compilers **********}

procedure sleep (msek: cardinal);

{$ifndef fpc}            {********** Compiler Delphi oder Kylix}

type TThread = class(classes.TThread)
{$ifdef Linux}
  FHandle : THandle;
{$endif}
 public
  procedure NotifyMainThread (Method: TThreadMethod);
  destructor destroy; override;
  constructor create(CreateSuspended: boolean); reintroduce;
end;

{$ifdef clx}
type TTimer = class(qextctrls.TTimer)
{$else}
type TTimer = class(extctrls.TTimer)
{$endif}
end;

type TCriticalSection = class (SyncObjs.TCriticalSection)
end;

function LockedDec (var Target: integer) : Integer;
function LockedINc (var Target: integer) : Integer;

{$else}                  {********** Compiler FPC **********}

type
    TFormMain = procedure of object;

{$include threadh.pas}


type TApplication = Class(TComponent)
   private
    Terminated: boolean;
    Threads : TFPThreadsRecord;
    FWaitEventException : TObject;
    FThreadsMax : integer;
    FSemasMax: integer;
    FSemasInited: boolean;
    FSemasCount : integer;
    PSemaArray : ^ByteBool;
    FCSSemID: Integer;
    FCSKey: TKey;
    FMsgQsMax: integer;
    FMsgQsInited: boolean;
    FMsgQsCount : integer;
    PMsgQArray : ^ByteBool;
    procedure WaitEvent;
    procedure SetThreadsMax (n: integer);
    procedure SetSemasMax (n: integer);
   public
    Handle: THandle;
    main: TFormMain;
    constructor create (AOwner: TComponent); override;
    procedure Initialize;
    function ThreadCount: integer;
    procedure CreateForm;
    procedure Run;
    procedure terminate;
    function ExeName: string;
    property Threadsmax : integer read FThreadsMax write SetThreadsMax;
    property Semassmax : integer read FSemasMax write SetSemasMax;
end;



type TFormClass = class (TApplication)
  private
   FOnCreate: TNotifyEvent;
  public
   constructor Create(AOWner: TComponent); override;
   procedure Formcreate (Sender: TObject); virtual; abstract;
   property OnCreate: TNotifyEvent write FOnCreate;
end;

type

  TTimerThread = class (TThread)
    Timer : TComponent;
    procedure execute; override;
  end;



  TTimer = class (TComponent)
  private
   FEnabled: boolean;
   FHandle: THandle;
   FInterval: integer;
   FThread : TTimerThread;
   FOnTimer: TNotifyEvent;
   procedure SetEnabled (En: boolean);
   procedure DoOnTimer;
  public
   property Enabled : boolean read fenabled write SetEnabled;
   property Handle : THandle read FHandle;
   property Interval : integer read FInterval write FInterval;
   property OnTimer: TNotifyEvent read FOnTimer write FOnTimer;
   constructor create (AOwner: TComponent); override;
   destructor destroy; override;
end;

TCriticalSection = class (TObject)
   FHandle : THandle;
  public
   procedure Acquire;
   constructor Create;
   destructor destroy; override;
   procedure Enter;
   procedure Leave;
   procedure Release;
   property Handle: THandle read FHandle;

end;

var
    Application : TApplication;

var
    FPFormList : TList;

{ TThreadList class }

type
  TThreadList = class
  private
    FList: TList;
    FLock: TCriticalSection;
    FDuplicates: TDuplicates;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Add(Item: Pointer);
    procedure Clear;
    function  LockList: TList;
    procedure Remove(Item: Pointer);
    procedure UnlockList;
    property Duplicates: TDuplicates read FDuplicates write FDuplicates;
  end;


{$endif}

                         {********** all compilers **********}
type
TSemaphore = class (TComponent)
  private
{$ifndef fpc} {Compiler Delphi oder Kylix}
   FActivity : integer;
   FEvent : TEvent;
{$else}
   FHandle : THandle;
{$endif}
   function GetCount: Integer;
  public
   constructor create (AOwner: TComponent); override;
   destructor destroy; override;
   procedure IncAndFree;
   procedure DecAndWait;
   procedure DecNoWait;
   procedure ResetEvent; virtual;
   procedure SetEvent;   virtual;
   procedure waitfor;
   property Count: integer read GetCount;
end;

{$ifdef FPC}
Type
TWaitResult = (wrSignaled, wrTimeout, wrAbandoned, wrError);
TEvent = class (TSemaphore)   //simulates Delphi TEvent
  public
   constructor Create(EventAttributes: Pointer;
                      ManualReset, InitialState: Boolean;
                      const Name: string);
   function WaitFor (TimeOut: DWord): TWaitResult;
end;
{$endif}



Type
TMessageQueue = class (TComponent)
  private
{$ifndef fpc} {Compiler Delphi oder Kylix}
   FEvent : TEvent;
   FList  : TThreadList;
   function GetCount: Integer;
{$else}
  FHandle: THandle;
  FCount: Integer;
{$endif}
  public
   constructor create (AOwner: TComponent); override;
   destructor destroy; override;
   procedure SendCardinal(V: Cardinal);
   procedure SendPointer(V: Pointer);
   function ReceiveCardinal: Cardinal;
   function ReceivePointer: Pointer;
{$ifndef fpc} {Compiler Delphi oder Kylix}
   property Count: integer read GetCount;
{$else}
   property Count: integer read FCount;
{$endif}
   property AsCardinal : Cardinal read ReceiveCardinal write SendCardinal;
   property AsPointer :  Pointer  read ReceivePointer  write SendPointer;
end;



{********** end Interface definitions **********}



implementation

{********** Implementation uses **********}


{$ifndef fpc}            {Compiler Delphi oder Kylix}
{$ifndef Linux}          {Compiler Delphi}
{$else}                  {Compiler Kylix}
uses libc;
{$endif}
{$else}                  {Compiler Free Pascal}
uses linux, strutils, sysutils, errors;
{$ASMMODE INTEL}
{$endif}





{********** end Implementation uses **********}


{********** Implementation definitions **********}

function LockedDec (var Target: integer) : Integer; assembler;
asm
	mov	ecx, Target
	mov     eax, -1
	lock xadd [ecx], eax
	dec	eax
end;


function LockedInc (var Target: integer) : Integer; assembler;
asm
	mov	ecx, Target
	mov     eax, 1
	lock xadd [ecx], eax
	inc	eax
end;


procedure TMessageQueue.SendCardinal(V: Cardinal);
begin
  SendPointer(Pointer(V));
end;

function TMessageQueue.ReceiveCardinal: Cardinal;
begin
  Result := Integer(ReceivePointer);
end;



{$ifndef fpc}                {********** Compiler Delphi oder Kylix **********}

{$ifndef LINUX}              {********** Compiler Delphi **********}

procedure sleep (msek: cardinal);
begin
{$ifdef clx}
  SysUtils.sleep(msek);
{$else}
  Windows.sleep(msek);
{$endif}
end;

{$else}                       {********** Compiler Kylix **********}


procedure sleep (msek: cardinal);
var req, rem : timespec;
    res : Integer;
begin
  req.tv_sec :=  msek div 1000;
  req.tv_nsec:= (msek mod 1000) * 1000000;
  repeat
     res := nanosleep (req,@rem);
     req := rem;
  until res >= 0;
//  writeln(res)
end;

{$endif}                       {********** Compiler Delphi or Kylix **********}

procedure TThread.NotifyMainThread (Method: TThreadMethod);
begin
   synchronize (Method);
end;

constructor TThread.create (CreateSuspended: boolean);
begin
  inherited;
{$ifdef Linux}
  FHandle := GetPID;
{$endif}
end;

destructor TThread.destroy;
begin
{$ifndef linux}
  TerminateThread (Handle, 0);
{$else}
//    Kill(FHandle, SIGKILL);          //does not work this way ?!?!?!
{$endif}
  inherited;
end;



constructor TSemaphore.Create (AOwner: TComponent);
var SNAme: string;
    i : integer;
begin
  inherited create (AOwner);
  randomize;
  i:=random (High(Integer));
  SNAME := ClassName + '-' + inttohex(i,8);
  FEvent := TEvent.Create(NIL, False, False, SName);
  FActivity:= 0;
end;

destructor TSemaphore.destroy;
begin
   FEvent.Free;
   inherited;
end;

procedure TSemaphore.IncAndFree;
begin
  LockedInc (FActivity);
  FEvent.SetEvent;
end;

procedure TSemaphore.DecAndWait;
begin
  while LockedDec (FActivity) < 0 do begin
    LockedInc (FActivity);
    FEvent.WaitFor(High(Cardinal));
  end;
end;

procedure TSemaphore.DecNoWait;
begin
  LockedDec (FActivity);
end;

procedure TSemaphore.ResetEvent;
begin
  FActivity := 0;
end;

procedure TSemaphore.SetEvent;
begin
  FActivity := 1;
  FEvent.SetEvent;
end;

procedure TSemaphore.WaitFor;
begin
  while FActivity <= 0 do begin
    FEvent.WaitFor(high(Cardinal));
  end;
end;

function TSemaphore.GetCount: Integer;
begin
  Result := FActivity;
end;



constructor TMessageQueue.create (AOwner: TComponent);
var SNAme: string;
    i : integer;
begin
  inherited create (AOwner);
  FList := TThreadList.Create;
  FList.Duplicates := dupAccept;
  randomize;
  i:=random(High(Integer));
  SNAME := ClassName + '-' + inttohex(i,8);
  FEvent := TEvent.Create(NIL, False, False, SName);
end;

destructor TMessageQueue.destroy;
begin
  FList.Free;
  FEvent.Free;
  inherited;
end;

function TMessageQueue.GetCount: Integer;
var L: TList;
begin
  L := FList.LockList;
  Result := L.Count;
  FList.UnlockList;
end;

procedure TMessageQueue.SendPointer(V: Pointer);
begin
  FList.Add(V);
  FEvent.SetEvent;
end;

function TMessageQueue.ReceivePointer: Pointer;
var L: TList;
begin
  L:=FList.LockList;
  while L.Count = 0 do begin
    FList.UnlockList;
    FEvent.WaitFor(High(Cardinal));
    L:=FList.LockList;
  end;
  Result := L[0];
  L.Delete(0);
  FList.UnlockList;
end;




{$else}                  {********** Compiler FPC **********}

procedure sleep (msek: cardinal);
var req, rem : timespec;
    res : Integer;
begin
  req.tv_sec :=  msek div 1000;
  req.tv_nsec:= (msek mod 1000) * 1000000;
  repeat
     res := nanosleep (req,rem);
     req := rem;
  until res >= 0;
end;


type TPSemUn = ^TSemUn;

procedure InitApp;
begin
   FPFormList := TList.create;
   Application := TApplication.Create (NIL);
end;





{$include thread.pas}




constructor TApplication.create (AOwner: TComponent);
const SemasMax = 10;
      MsgQsMax = 10;
begin
  inherited;
  Handle := GetPID;
  FThreadsMax := FPThreadsMax;
  FSemasMax := SemasMax;
  FMsgQsMax := MsgQsMax;
end;


procedure TApplication.Initialize;
begin
//   writeln ('          *** TApplication.Initialize ***');
end;

function TApplication.Threadcount: integer;
var   hp  : PFPThreadRec;
begin
   hp := Application.Threads.Root;
   Result := 0;
   while (hp <> NIL) do begin
//                           writeln ('          *** hp (1) = ', inttohex(integer(hp),8), ' ***');
       inc (Result);
//                           writeln ('          *** hp/Killed = ', inttohex(integer(hp),8), '/',Application.Threads,Killed, ' ***');
       hp := hp^.next;
//                           writeln ('          *** hp (2) = ', inttohex(integer(hp),8), ' ***');
   end;
end;

procedure TApplication.WaitEvent;
var
   res : Boolean;
   msg : TFPThreadMessage;    //PMsgBuf;
   sem : TSemBuf;
   ThisHandle2 : THandle;
   ThisHandle  : THandle;
const msgflg = 0;
      MType = FPMsgSynchonize;
begin
  Terminated := false;
  FWaitEventException := NIL;
  repeat
     res := msgrcv(Application.Threads.MsgID, @msg, sizeof(TFPThreadMessage), MType, msgflg);
//        writeln ('          >>> ', res, ' ', msg.MType, ' ', inttohex (integer(msg.Sender), 8), ' ', sys_errlist[IPCError], ' (',IPCError, ') <<<');
     if res then begin
       case msg.MType of
          FPMsgSynchonize : begin
            if not assigned (msg.Sender) then begin
               Terminated := true;
             end else begin
//                      writeln ('          *** synchronizing Thread ID:        ', inttohex (integer(msg.Sender),8), '(', msg.Sender.FThreadID, ') msg.Sender.M: ', inttohex(integer(msg.Sender),8), '.', inttohex(integer(@msg.Sender.FMethod),8), ' ***');
              ThisHandle  := msg.Sender.FHandle;
              ThisHandle2 := msg.Sender.FHandle2;  // in case a future version of Linux reuses the PID instantly
              sem.sem_num := msg.Sender.FThreadID;
              sem.sem_op  := 1;
              sem.sem_flg := IPC_NOWAIT;
              if assigned (msg.Sender.FMethod) then msg.Sender.FMethod;
//                  writeln ('          *** Application.Threads.Freed: ', Application.Threads.Freed, '  ***');
              if Application.Threads.Freed then begin  //the user might have freed the thread
                 Application.Threads.Freed := IsThisThreadFreed (ThisHandle, ThisHandle2);
              end;
              if not Application.Threads.Freed then begin
                res := semop  (Application.Threads.SemID, @sem, 1);    // inc Semaphore
                if not res then begin
                  if IPCError <> 4 then begin  // "intterupted System Call" seems to be not critical
                    raise (FWaitEventException);
//                    writeln ('          ??? Synchronize semop 2 IPCError: ', sys_errlist[IPCError], ' (', IPCError,') ???');
                  end;
                end;
              end;
              Application.Threads.Freed := false; //no need to synchronize as Freed is only changed in the Mainthread
            end;
          end;
       end;
      end else begin
        if IPCError <> 4 then begin  // "intterupted System Call" seems to be not critical
          raise (FWaitEventException);
//          writeln ('          ??? ', res, ' ', msg.MType, ' ', msg.Sender.ThreadID, ' ', sys_errlist[IPCError], ' (', IPCError , ') ???');
        end;
     end;
  until Terminated  or (Application.Threads.Root = NIL);
//        writeln ('          *** End Message-Loop *** ');
end;


procedure TApplication.CreateForm;
var f: TFormClass;
    i: integer;
begin
//     writeln ('--- TApplication.CreateForm Start ---');
   for i := 0 to FPFormList.Count-1 do begin
//     writeln ('--- TApplication.CreateForm ' + inttostr(i) + ' ---');
     F := TFormClass(FPFormList.Items[I]);
     if assigned (F.FOnCreate) then F.FOnCreate (self);
   end;
end;



procedure TApplication.Run;
begin
//   writeln ('--- TApplication.Run  ---');
  if assigned (main) then begin
//    writeln ('--- starting main ---');
    main;
  end;
  WaitEvent;
end;

procedure TApplication.Terminate;
var res : boolean;
   msg : TFPThreadMessage;    //PMsgBuf;
const msgflg = 0;
begin
//        writeln ('          *** Application Terminate auf PID ' + inttostr (GetPID) + ' ***');
  Application.Threads.Freed := true;
  msg.MType := FPMsgSynchonize;
  msg.Sender:= NIL;
  res := msgsnd (Application.Threads.MsgID, @msg, sizeof(msg), msgflg);
  if not res then begin
//     writeln ('          *** Application.Terminate IPCError: ', sys_errlist[IPCError], ' (', IPCError,') ***');
  end;
end;

procedure TApplication.SetThreadsMax (n: integer);
var FThreadsMaxException: TObject;
begin
  FThreadsMaxException := NIL;
  if Threads.ThreadsInited then begin
    raise (FThreadsMaxException);
   end else begin
    FThreadsMax := n;
  end;
end;

procedure TApplication.SetSemasMax (n: integer);
var FSemasMaxException: TObject;
begin
  FSemasMaxException := NIL;
  if FSemasInited then begin
    raise (FSemasMaxException);
   end else begin
    FSemasMax := n;
  end;
end;

function TApplication.ExeName: string;
begin
  Result := Paramstr(0);
end;




constructor TFormClass.Create(AOWner: TComponent);
begin
   inherited;
//   writeln ('--- TFormClass.Create ---');
end;




procedure TTimerThread.Execute;
begin
//  setitimer (100);
//  writeln ('          ### Timer Execute ThreadID/Handle/Handle2/', ThreadID, '/', Handle,   '/', Handle2, ' ###');
  while not terminated do begin
     sleep ((Timer as TTimer).FInterval);
//     write ('.');
     if not terminated then synchronize ((Timer as TTimer).DoOnTimer);
  end;
//     write ('!');
end;

procedure TTimer.DoOnTimer;
begin
  if assigned (FOnTimer) then FOnTimer(self)
end;

constructor TTimer.create (AOwner: TComponent);
begin
//  writeln ('          ### TTimer.create ###');
  inherited;
//  writeln ('          ### TTimer.create after inherited ###');
  FThread       := TTimerThread.create (true);
//  writeln ('          ### TTimer.create after TTimerThread.create, Fthread = ', inttohex(integer(FThread), 8), '  ###');
//  writeln ('          ### Fthread.ThreadID = ', Fthread.ThreadID);
  FThread.Timer := self;
  FHandle       := FThread.Handle;  // PID
end;

destructor TTimer.destroy;
begin
//  writeln ('          ### Timer destroy FThread.Terminate ###');
  FThread.free;
  inherited;
end;

procedure TTimer.SetEnabled (En: boolean);
begin
   FEnabled := En;
   FThread.suspended := not En;
//   writeln ('          ### Timer.Eneabled ', en, ' ###');
end;

Const
  CSDir = '.';
  CSChar= char (ord('S') + $80);


function GetNextUnusedSema: THandle;
var S: ^Bytebool;
begin
  Result := 0;
  S := Application.PSemaArray;
  While S^ do begin
    inc (Result);
    inc (S);
  end;
  S^ := true;
end;

procedure SetUnusedSema (Handle: THandle);
var S: ^Bytebool;
begin
  S := Application.PSemaArray;
  inc (S, Handle);
  S^ := false;
end;

procedure InitializeSemas;
const oct666 =  438;
var FSemasInitException : TObject;
    SemUn : TSemUn;
    res: LongInt;
begin
  getmem   (Application.PSemaArray,  Application.FSemasCount);
  fillchar (Application.PSemaArray^, Application.FSemasCount, bytebool(False));
  FSemasInitException := NIL;
  Application.FCSKey := ftok (CSDir, CSChar);
  Application.FCSSemID := semget(Application.FCSKey,Application.FSemasMax, IPC_CREAT or IPC_EXCL or oct666);     // create, must not exist
//     writeln ('          --- semget FCSKey/Application.TFCSSemID/IPCError ' + inttohex(Application.FCSKey,8) + '/' + inttohex(Application.FCSSemID,8) + '/' + sys_errlist[IPCError] + ' ***');
  if Application.FCSSemID = -1 then begin
    Application.FCSSemID := semget(Application.FCSKey,Application.FSemasMax, oct666);     // attach
//       writeln ('          --- semget Application.FCSKey/Application.FCSSemID/IPCError ' + inttohex(Application.FCSKey,8) + '/' + inttohex(Application.FCSSemID,8) + '/' + sys_errlist[IPCError] + ' ***');
    res := semctl(Application.FCSSemID, Application.FSemasMax, IPC_RMID, SemUn);   // exists
//       writeln ('          --- semctl IPC_RMID res/IPCError ', res, '/', sys_errlist[IPCError] + ' ***');
    if res = -1 then begin
      raise (FSemasInitException);
//      halt;
    end;
    Application.FCSSemID := semget (Application.FCSKey, Application.FSemasMax, IPC_CREAT or oct666);     // create if notexist
//       writeln ('          --- semget Application.FCSKey/Application.FCSSemID/IPCError ' + inttohex(Application.FCSKey,8) + '/' + inttohex(Application.FCSSemID,8) + '/' + sys_errlist[IPCError] + ' ***');
  end;
  Application.FSemasInited := true;
end;

procedure FinalizeSemas;
var
  SemUn : TSemUn;
//  Res: LongInt;
begin
  {res:=} semctl(Application.FCSSemID, Application.FSemasMax, IPC_RMID, SemUn);
//     writeln ('          --- semctl IPC_RMID res/IPCError ', res, '/', sys_errlist[IPCError] + ' ***');
  Application.FSemasInited := false;
  freemem (Application.PSemaArray);
end;

procedure TCriticalSection.Acquire;
begin
  Enter;
end;

constructor TCriticalSection.Create;
var
  SemUn : TSemUn;
//  res: integer;
  FSemasMaxException: TObject;
begin
  inherited;
  if not Application.FSemasInited then InitializeSemas;
  if Application.FSemasCount >= Application.FSemasMax then begin
    FSemasMaxException := NIL;
    raise (FSemasMaxException);
  end;
  inc (Application.FSemasCount);
  FHandle := GetNextUnusedSema;
  SemUn.val := 1;
  {res:=} semctl(Application.FCSSemID, FHandle, SETVAL, SemUn);
//       writeln ('          --- semctl SETVAL res/IPCError ', res, '/', sys_errlist[IPCError] + ' ***');

end;

destructor TCriticalSection.destroy;
begin
  SetUnusedSema (FHandle);
  dec (Application.FSemasCount);
  if Application.FSemasCount = 0 then FinalizeSemas;
  inherited;
end;

procedure TCriticalSection.Enter;
var
  sem : TSemBuf;
  res : Boolean;
  FCriticalSectionsInitException : TObject;
begin
  FCriticalSectionsInitException := NIL;
  sem.sem_num := handle;
  sem.sem_op  := -1;
  sem.sem_flg := 0;
  res := semop  (Application.FCSSemID, @sem, 1);    // dec Semaphore and maybe wait
  if not res then begin
    raise (FCriticalSectionsInitException);
//       writeln ('          ??? ENTER semop 1 IPCError: ', sys_errlist[IPCError], ' ???');
  end;
end;

procedure TCriticalSection.Leave;
var
  sem : TSemBuf;
  res : Boolean;
  FCriticalSectionException : TObject;
begin
  sem.sem_num := handle;
  sem.sem_op  := 1;
  sem.sem_flg := 0;
  res := semop  (Application.FCSSemID, @sem, 1);    // inc Semaphore
  if not res then begin
    FCriticalSectionException := NIL;
    raise (FCriticalSectionException);
//       writeln ('          ??? ENTER semop 1 IPCError: ', sys_errlist[IPCError], ' ???');
  end;
end;                        

procedure TCriticalSection.Release;
begin
  Leave;
end;

constructor TSemaphore.Create (AOwner: TComponent);
var
  SemUn : TSemUn;
//  res: integer;
  FSemasMaxException: TObject;
begin
  inherited create (AOwner);
  if not Application.FSemasInited then InitializeSemas;
  if Application.FSemasCount >= Application.FSemasMax then begin
    FSemasMaxException := NIL;
    raise (FSemasMaxException);
  end;
  inc (Application.FSemasCount);
  FHandle := GetNextUnusedSema;
  SemUn.val := 0;
  {res:=} semctl(Application.FCSSemID, FHandle, SETVAL, SemUn);
//       writeln ('          --- TSemaphore.Create: semctl SETVAL res/IPCError ', res, '/', sys_errlist[IPCError] + ' ***');

end;

destructor TSemaphore.destroy;
begin
  SetUnusedSema (FHandle);
  dec (Application.FSemasCount);
  if Application.FSemasCount = 0 then FinalizeSemas;
  inherited;
end;

procedure TSemaphore.IncAndFree;
var
  sem : TSemBuf;
  res : Boolean;
  FSemaException : TObject;
begin
  sem.sem_num := FHandle;
  sem.sem_op  := 1;
  sem.sem_flg := 0;
  res := semop  (Application.FCSSemID, @sem, 1);    // inc Semaphore
  if not res then begin
    FSemaException := NIL;
    raise (FSemaException);
//       writeln ('          ??? ENTER semop 1 IPCError: ', sys_errlist[IPCError], ' ???');
  end;
end;

procedure TSemaphore.DecAndWait;
var
  sem : TSemBuf;
  res : Boolean;
  FSemaException : TObject;
begin
  sem.sem_num := FHandle;
  sem.sem_op  := -1;
  sem.sem_flg := 0;
  res := semop  (Application.FCSSemID, @sem, 1);    // dec Semaphore and maybe wait
  if not res then begin
    FSemaException := NIL;
    raise (FSemaException);
//       writeln ('          ??? ENTER semop 1 IPCError: ', sys_errlist[IPCError], ' ???');
  end;
end;

procedure TSemaphore.DecNoWait;
var
  sem : TSemBuf;
  res : Boolean;
  FSemaException : TObject;
begin
  sem.sem_num := FHandle;
  sem.sem_op  := -1;
  sem.sem_flg := IPC_NOWAIT;
  res := semop  (Application.FCSSemID, @sem, 1);    // dec Semaphore and don't wait
  if not res then begin
    FSemaException := NIL;
    raise (FSemaException);
//       writeln ('          ??? ENTER semop 1 IPCError: ', sys_errlist[IPCError], ' ???');
  end;
end;

procedure TSemaphore.ResetEvent;
var
//  res : Boolean;
  SemUn : TSemUn;
begin
  SemUn.val := 0;
  {res:=} semctl(Application.FCSSemID, FHandle, SETVAL, SemUn);
//       writeln ('          --- TSemaphore.Create: semctl SETVAL res/IPCError ', res, '/', sys_errlist[IPCError] + ' ***');
end;

procedure TSemaphore.SetEvent;
var
//  res : Boolean;
  SemUn : TSemUn;
begin
  SemUn.val := 1;
  {res:=} semctl(Application.FCSSemID, FHandle, SETVAL, SemUn);
//       writeln ('          --- TSemaphore.Create: semctl SETVAL res/IPCError ', res, '/', sys_errlist[IPCError] + ' ***');
end;

procedure TSemaphore.Waitfor;
begin
  DecAndWait;
end;

function TSemaphore.GetCount: Integer;
var PSemUn : ^TSemUn;
begin
  PSemUn := NIL;
  Result := semctl(Application.FCSSemID, FHandle, GetVAL, PSemUn^);
end;


procedure InitializeMsgQs;
begin
  getmem   (Application.PMsgQArray,  Application.FMsgQsCount);
  fillchar (Application.PMsgQArray^, Application.FMsgQsCount, bytebool(False));
  Application.FMsgQsInited := true;
end;

procedure FinalizeMsgQs;
begin
  Application.FMsgQsInited := false;
  freemem (Application.PMsgQArray);
end;

function GetNextUnusedMsgQ: THandle;
var S: ^Bytebool;
begin
  Result := 0;
  S := Application.PMsgQArray;
  While S^ do begin
    inc (Result);
    inc (S);
  end;
  S^ := true;
end;

procedure SetUnusedMsgQ (Handle: THandle);
var S: ^Bytebool;
begin
  S := Application.PMsgQArray;
  inc (S, Handle);
  S^ := false;
end;

constructor TMessageQueue.create (AOwner: TComponent);
var
    FMsgQsMaxException : TObject;
begin
  inherited;
//     Writeln ('          --- TMessageQueue.Create ---');
  if not Application.FMsgQsInited then InitializeMsgQs;
  if Application.FMsgQsCount >= Application.FMsgQsMax then begin
    FMsgQsMaxException := NIL;
    raise (FMsgQsMaxException);
  end;
  inc (Application.FMsgQsCount);
  FHandle := GetNextUnusedMsgQ;
  FCount := 0;
end;

destructor TMessageQueue.Destroy;
begin
//     Writeln ('          --- TMessageQueue.Destroy ---');
  SetUnusedMsgQ (FHandle);
  dec (Application.FMsgQsCount);
  if Application.FMsgQsCount = 0 then FinalizeMsgQs;
  inherited;
end;

type TMsgQMessage = record
    MType  : Longint;
    V      : Pointer;
end;

const MsqQMessageType = $1000;
      TMsgQMessageSize = sizeof(TMsgQMessage)-sizeof(Longint);

procedure TMessageQueue.SendPointer(V: Pointer);
var
   FMsgQException : TObject;
   msg : TMsgQMessage;    //PMsgBuf;
   res : Boolean;
const msgflg = 0;
begin
    msg.MType := MsqQMessageType + FHandle;
    msg.V:= V;
    inc (FCount);
    res := msgsnd (Application.Threads.MsgID, @msg,    TMsgQMessageSize, msgflg);
//        writeln ('          >>> send ', res, ' ', msg.MType, ' ', inttohex (integer(msg.V), 8), ' ', sys_errlist[IPCError], ' (',IPCError, ') :', FCount, ' <<<');
    if not res then begin
      FMsgQException := NIL;
      raise (FMsgQException);
//       writeln ('          ??? SendPointer msgsnd IPCError: ', sys_errlist[IPCError], ' ???');
    end;
end;

function TMessageQueue.ReceivePointer: Pointer;
var
   FMsgQException : TObject;
   msg : TMsgQMessage;    //PMsgBuf;
   res : Boolean;
const msgflg = 0;
begin
   res := msgrcv(Application.Threads.MsgID, @msg, sizeof(TMsgQMessage), MsqQMessageType+FHandle, msgflg);
//        writeln ('          >>> rec', res, ' ', msg.MType, ' ', inttohex (integer(msg.V), 8), ' ', sys_errlist[IPCError], ' (',IPCError, ') :', FCount, ' <<<');
   if not res then begin
      FMsgQException := NIL;
      raise (FMsgQException);
//       writeln ('          ??? ReceivePointer msgsnd IPCError: ', sys_errlist[IPCError], ' ???');
   end;
   Result := msg.V;
   dec (FCount);
end;

constructor TEvent.Create(EventAttributes: Pointer;
                   ManualReset, InitialState: Boolean;
                   const Name: string);
begin
   inherited create (Application);
   if InitialState then begin
     IncAndFree;
   end;
end;

function TEvent.Waitfor (Timeout: DWORD): TWaitResult;
begin
  DecAndWait;
  Result := wrSignaled;
end;

{ TThreadList }

constructor TThreadList.Create;
begin
  inherited Create;
  FLock := TCriticalSection.Create;
  FList := TList.Create;
  FDuplicates := dupIgnore;
end;

destructor TThreadList.Destroy;
begin
  LockList;    // Make sure nobody else is inside the list.
  try
    FList.Free;
    inherited Destroy;
  finally
    UnlockList;
    FLock.free;
  end;
end;

const
  SDuplicateItem : string = 'List does not allow duplicates ($0%x)';


procedure TThreadList.Add(Item: Pointer);
begin
  LockList;
  try
    if (Duplicates = dupAccept) or
       (FList.IndexOf(Item) = -1) then
      FList.Add(Item)
    else if Duplicates = dupError then
      FList.Error(SDuplicateItem, Integer(Item));
  finally
    UnlockList;
  end;
end;

procedure TThreadList.Clear;
begin
  LockList;
  try
    FList.Clear;
  finally
    UnlockList;
  end;
end;

function  TThreadList.LockList: TList;
begin
  FLock.Enter;
  Result := FList;
end;

procedure TThreadList.Remove(Item: Pointer);
begin
  LockList;
  try
    FList.Remove(Item);
  finally
    UnlockList;
  end;
end;

procedure TThreadList.UnlockList;
begin
  FLock.Leave;
end;




{$endif} {FPC}

{********** end Implementation definitions **********}



{********** Imnitialization definitions **********}

{$ifdef FPC}                 {********** Compiler FPC **********}
initialization
begin
//  writeln ('--- spcompatibility initialization ---');
  InitApp;
//  writeln ('--- after InitApp ---');
end;

{$endif}


end. tlist
